// Boost. Range Template Library (rtl)
//
// Copyright 2003 John Torjo (john@torjo.com) and Matthew Wilson (matthew@synesis.com.au)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
// See http://www.boost.org for updates, documentation, and revision history.

#ifndef BOOST_RTL_ALGO_HPP_INCLUDED
#define BOOST_RTL_ALGO_HPP_INCLUDED

#include <algorithm>
#include <iterator>
#include <iosfwd>
#include <boost/config.hpp>
#include <boost/rtl/rtl.hpp>
#include <boost/rtl/priv/traits.hpp>

/*
    Notes:
    1. For the purpose of algorithms, EVERY container IS a range.

    2. All predicates are passed by value, since VC6 does not hadle 'const Pred&' well,
    when they are functions.
    
    3. At this time (14th Nov 2003), all STL algorithms are implemented, except for:
    - partial_sort_copy
    - mismatch


// FIXME - implement <numeric> algorithms as well
*/


// rtl = Range Template Library
namespace boost { namespace rtl { 


template< class value_type, class char_type, class traits_type>
irange< ::std::istream_iterator< value_type, char_type, traits_type> > istream_range( ::std::basic_istream<char_type,traits_type> & s) {
    typedef ::std::istream_iterator< value_type, char_type, traits_type> it_type;
    typedef irange< it_type > range_type;
    return range_type( it_type(s), it_type() );
}


namespace rng {



// easy access to .find() from associative containers
//
// Example:
// if ( rng::coll_find(m_cached_days, current_day) ) {
// }
//
// equivalent to:
// if ( m_cached_days.find(current_day) != m_cached_days.end() ) {
// }
template< class container, class value_type> inline irange<typename container::iterator> 
coll_find( container & c, const value_type & val) {
    typedef irange<typename container::iterator> range_type;
    return range_type( c.find(val), c.end() );
}

template< class container, class value_type> inline crange<const container> 
coll_find( const container & c, const value_type & val) {
    return crange<const container>( c.find(val), c.end() );
}



// FIXME - explain how it is to be used!!!
template<class container, class r> inline void
erase( container & c, const crange<r> & rng) {
    c.erase( rng.begin(), rng.end() );
}








// FIXME - allow concatenation of multiple ranges/containers of the same type.
//         TOTHINK if it's really needed




// helper - for passing an iterator instead of a range as a second parameter,
//          to an algorithm that needs two ranges (example: mismatch)
//
// You can have 
// - rng::mismatch( rng1, rng2); // passing two ranges
// - rng::mismatch( rng, i_(iterator) ); // passing the second param as an iterator
// - rng::mismatch( rng, i_param(iterator) ); // same as above, but more descriptive
template< class it>
struct i_param_t {
    typedef it iterator;
    i_param_t( iterator first) : m_first(first) {}
    iterator begin() const { return m_first;}

    // provided for copy_backward, which if a range is the second param, will copy starting from its end
    iterator end() const { return m_first;}
private:
    iterator m_first;
};

// shorter form: 'i_'
template<class it> inline i_param_t<it> i_( it val) { return i_param_t<it>(val); }
// longer and more descriptable/readable form 'i_param'
template<class it> inline i_param_t<it> i_param( it val) { return i_param_t<it>(val); }






///////////////////////////////////////////////////////////////////////////////
// Macros for algorithms that DONOT return ranges


#define BOOST_RTL_RNG_ALGO10_RAW(name, ret_type, val1_type, val1_param, return_keyw) \
template<class val1_type> inline ret_type                                   \
name(val1_param val1) {                                                     \
    return_keyw ::std::name( val1.begin(), val1.end() );                \
}

#define BOOST_RTL_RNG_ALGO11_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, return_keyw) \
template<class val1_type, class val2_type> inline ret_type                      \
name(val1_param val1, val2_param val2) {                                        \
    return_keyw ::std::name( val1.begin(), val1.end(), val2 );              \
}

#define BOOST_RTL_RNG_ALGO12_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param, return_keyw) \
template<class val1_type, class val2_type, class val3_type> inline ret_type             \
name(val1_param val1, val2_param val2, val3_param val3) {                               \
    return_keyw ::std::name( val1.begin(), val1.end(), val2, val3 );                \
}

#define BOOST_RTL_RNG_ALGO13_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, return_keyw) \
template<class val1_type, class val2_type, class val3_type, class val4_type> inline ret_type        \
name(val1_param val1, val2_param val2, val3_param val3, val4_param val4) {                          \
    return_keyw ::std::name( val1.begin(), val1.end(), val2, val3, val4 );                      \
}



#define BOOST_RTL_RNG_ALGO20_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, return_keyw) \
template<class val1_type, class val2_type> inline ret_type                              \
name(val1_param val1, val2_param val2) {                                                \
    return_keyw ::std::name( val1.begin(), val1.end(), val2.begin(), val2.end() );  \
}

#define BOOST_RTL_RNG_ALGO21_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param, return_keyw) \
template<class val1_type, class val2_type, class val3_type> inline ret_type                     \
name(val1_param val1, val2_param val2, val3_param val3) {                                       \
    return_keyw ::std::name( val1.begin(), val1.end(), val2.begin(), val2.end(), val3 );    \
}

#define BOOST_RTL_RNG_ALGO22_RAW(name, ret_type, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, return_keyw) \
template<class val1_type, class val2_type, class val3_type, class val4_type> inline ret_type                \
name(val1_param val1, val2_param val2, val3_param val3, val4_param val4) {                                  \
    return_keyw ::std::name( val1.begin(), val1.end(), val2.begin(), val3, val4 );          \
}



#ifndef BOOST_RTL_WORKAROUND_VC6

#define BOOST_RTL_RNG_ALGO10(name, ret_type) \
    BOOST_RTL_RNG_ALGO10_RAW( name, ret_type, r, const r&, return) \
    BOOST_RTL_RNG_ALGO10_RAW( name, ret_type, r, r&, return)

#define BOOST_RTL_RNG_ALGO11(name, ret_type, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_RAW( name, ret_type, r, const r&, val2_type, val2_param, return) \
    BOOST_RTL_RNG_ALGO11_RAW( name, ret_type, r, r&, val2_type, val2_param, return)

#define BOOST_RTL_RNG_ALGO12(name, ret_type, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_RAW( name, ret_type, r, const r&, val2_type, val2_param, val3_type, val3_param, return) \
    BOOST_RTL_RNG_ALGO12_RAW( name, ret_type, r, r&, val2_type, val2_param, val3_type, val3_param, return)

#define BOOST_RTL_RNG_ALGO13(name, ret_type, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_RAW( name, ret_type, r, const r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, return) \
    BOOST_RTL_RNG_ALGO13_RAW( name, ret_type, r, r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, return)

#define BOOST_RTL_RNG_ALGO10_void(name) \
    BOOST_RTL_RNG_ALGO10_RAW( name, void, r, const r&, (void)) \
    BOOST_RTL_RNG_ALGO10_RAW( name, void, r, r&, (void))

#define BOOST_RTL_RNG_ALGO11_void(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_RAW( name, void, r, const r&, val2_type, val2_param, (void)) \
    BOOST_RTL_RNG_ALGO11_RAW( name, void, r, r&, val2_type, val2_param, (void))

#define BOOST_RTL_RNG_ALGO12_void(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_RAW( name, void, r, const r&, val2_type, val2_param, val3_type, val3_param, (void)) \
    BOOST_RTL_RNG_ALGO12_RAW( name, void, r, r&, val2_type, val2_param, val3_type, val3_param, (void))

#define BOOST_RTL_RNG_ALGO13_void(name, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_RAW( name, void, r, const r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, (void)) \
    BOOST_RTL_RNG_ALGO13_RAW( name, void, r, r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, (void))


#define BOOST_RTL_RNG_ALGO20(name, ret_type) \
    BOOST_RTL_RNG_ALGO20_RAW( name, ret_type, r1, r1&, r2, const r2&, return) \
    BOOST_RTL_RNG_ALGO20_RAW( name, ret_type, r1, r1&, r2, r2&, return) \
    BOOST_RTL_RNG_ALGO20_RAW( name, ret_type, r1, const r1&, r2, const r2&, return) \
    BOOST_RTL_RNG_ALGO20_RAW( name, ret_type, r1, const r1&, r2, r2&, return)

#define BOOST_RTL_RNG_ALGO21(name, ret_type, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_RAW( name, ret_type, r1, r1&, r2, const r2&, val3_type, val3_param, return) \
    BOOST_RTL_RNG_ALGO21_RAW( name, ret_type, r1, r1&, r2, r2&, val3_type, val3_param, return) \
    BOOST_RTL_RNG_ALGO21_RAW( name, ret_type, r1, const r1&, r2, const r2&, val3_type, val3_param, return) \
    BOOST_RTL_RNG_ALGO21_RAW( name, ret_type, r1, const r1&, r2, r2&, val3_type, val3_param, return)

#define BOOST_RTL_RNG_ALGO22(name, ret_type, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO22_RAW( name, ret_type, r1, r1&, r2, const r2&, val3_type, val3_param, val4_type, val4_param, return) \
    BOOST_RTL_RNG_ALGO22_RAW( name, ret_type, r1, r1&, r2, r2&, val3_type, val3_param, val4_type, val4_param, return) \
    BOOST_RTL_RNG_ALGO22_RAW( name, ret_type, r1, const r1&, r2, const r2&, val3_type, val3_param, val4_type, val4_param, return) \
    BOOST_RTL_RNG_ALGO22_RAW( name, ret_type, r1, const r1&, r2, r2&, val3_type, val3_param, val4_type, val4_param, return)

#define BOOST_RTL_RNG_ALGO20_void(name) \
    BOOST_RTL_RNG_ALGO20_RAW( name, void, r1, r1&, r2, const r2&, (void)) \
    BOOST_RTL_RNG_ALGO20_RAW( name, void, r1, r1&, r2, r2&, (void)) \
    BOOST_RTL_RNG_ALGO20_RAW( name, void, r1, const r1&, r2, const r2&, (void)) \
    BOOST_RTL_RNG_ALGO20_RAW( name, void, r1, const r1&, r2, r2&, (void))

#else

// workaround for VC6
// VC6 chokes when sees a function having two templated overloads, one const and one non-const

#define BOOST_RTL_RNG_ALGO10(name, ret_type) \
    BOOST_RTL_RNG_ALGO10_RAW( name, ret_type, r, r&, return)

#define BOOST_RTL_RNG_ALGO11(name, ret_type, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_RAW( name, ret_type, r, r&, val2_type, val2_param, return)

#define BOOST_RTL_RNG_ALGO12(name, ret_type, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_RAW( name, ret_type, r, r&, val2_type, val2_param, val3_type, val3_param, return)

#define BOOST_RTL_RNG_ALGO13(name, ret_type, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_RAW( name, ret_type, r, r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, return)

#define BOOST_RTL_RNG_ALGO10_void(name) \
    BOOST_RTL_RNG_ALGO10_RAW( name, void, r, r&, (void))

#define BOOST_RTL_RNG_ALGO11_void(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_RAW( name, void, r, r&, val2_type, val2_param, (void))

#define BOOST_RTL_RNG_ALGO12_void(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_RAW( name, void, r, r&, val2_type, val2_param, val3_type, val3_param, (void))

#define BOOST_RTL_RNG_ALGO13_void(name, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_RAW( name, void, r, r&, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param, (void))

#define BOOST_RTL_RNG_ALGO20(name, ret_type) \
    BOOST_RTL_RNG_ALGO20_RAW( name, ret_type, r1, r1&, r2, r2&, return)

#define BOOST_RTL_RNG_ALGO21(name, ret_type, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_RAW( name, ret_type, r1, r1&, r2, r2&, val3_type, val3_param, return)

#define BOOST_RTL_RNG_ALGO22(name, ret_type, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO22_RAW( name, ret_type, r1, r1&, r2, r2&, val3_type, val3_param, val4_type, val4_param, return)

#define BOOST_RTL_RNG_ALGO20_void(name) \
    BOOST_RTL_RNG_ALGO20_RAW( name, void, r1, r1&, r2, r2&, (void))

#endif // BOOST_RTL_WORKAROUND_VC6

// END OF Macros for algorithms that DONOT return ranges
///////////////////////////////////////////////////////////////////////////////






//////////////////////////////////////////////////////////////////////////////
// Macros for algorithms that return ranges

// note: ranges are always passed by reference

#define BOOST_RTL_RNG_ALGO10_EXT_RAW(name, val1_type, val1_param ) \
template<class val1_type> inline typename range_finder<val1_param>::range_type          \
name(val1_param& val1) {                                                                \
    typedef typename range_finder<val1_param>::range_type range_type;                   \
    return range_type( ::std::name( val1.begin(), val1.end() ), val1.end() );         \
}

#define BOOST_RTL_RNG_ALGO11_EXT_RAW(name, val1_type, val1_param, val2_type, val2_param ) \
template<class val1_type, class val2_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    return range_type( ::std::name( val1.begin(), val1.end(), val2 ), val1.end() );                   \
}

#define BOOST_RTL_RNG_ALGO12_EXT_RAW(name, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param ) \
template<class val1_type, class val2_type, class val3_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2, val3_param val3) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    return range_type( ::std::name( val1.begin(), val1.end(), val2, val3 ), val1.end() );                   \
}

// especially for equal_range
#define BOOST_RTL_RNG_ALGO11_EXT_RAW2(name, val1_type, val1_param, val2_type, val2_param ) \
template<class val1_type, class val2_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    typedef typename range_type::iterator iterator; \
    return ::boost::rtl::irange<iterator>( ::std::name( val1.begin(), val1.end(), val2 ) );                   \
}

// especially for equal_range
#define BOOST_RTL_RNG_ALGO12_EXT_RAW2(name, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param ) \
template<class val1_type, class val2_type, class val3_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2, val3_param val3) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    typedef typename range_type::iterator iterator; \
    return ::boost::rtl::irange<iterator>( ::std::name( val1.begin(), val1.end(), val2, val3 ) );                   \
}

#define BOOST_RTL_RNG_ALGO13_EXT_RAW(name, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param ) \
template<class val1_type, class val2_type, class val3_type, class val4_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2, val3_param val3, val4_param val4) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    return range_type( ::std::name( val1.begin(), val1.end(), val2, val3, val4 ), val1.end() );                   \
}

#define BOOST_RTL_RNG_ALGO20_EXT_RAW(name, val1_type, val1_param, val2_type, val2_param ) \
template<class val1_type, class val2_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                   \
    return range_type( ::std::name( val1.begin(), val1.end(), val2.begin(), val2.end() ), val1.end() );                   \
}

#define BOOST_RTL_RNG_ALGO21_EXT_RAW(name, val1_type, val1_param, val2_type, val2_param, val3_type, val3_param ) \
template<class val1_type, class val2_type, class val3_type> inline typename range_finder<val1_param>::range_type         \
    name(val1_param& val1, val2_param val2, val3_param val3) {                                                           \
    typedef typename range_finder<val1_param>::range_type range_type;                                                    \
    return range_type( ::std::name( val1.begin(), val1.end(), val2.begin(), val2.end(), val3 ), val1.end() );          \
}


#ifndef BOOST_RTL_WORKAROUND_VC6

#define BOOST_RTL_RNG_ALGO10_EXT(name) \
    BOOST_RTL_RNG_ALGO10_EXT_RAW( name, r, const r) \
    BOOST_RTL_RNG_ALGO10_EXT_RAW( name, r, r)

#define BOOST_RTL_RNG_ALGO11_EXT(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW( name, r, const r, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW( name, r, r, val2_type, val2_param)

#define BOOST_RTL_RNG_ALGO12_EXT(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW( name, r, const r, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW( name, r, r, val2_type, val2_param, val3_type, val3_param)

// especially for equal_range
#define BOOST_RTL_RNG_ALGO11_EXT2(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW2( name, r, const r, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW2( name, r, r, val2_type, val2_param)

// especially for equal_range
#define BOOST_RTL_RNG_ALGO12_EXT2(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW2( name, r, const r, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW2( name, r, r, val2_type, val2_param, val3_type, val3_param)

#define BOOST_RTL_RNG_ALGO13_EXT(name, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_EXT_RAW( name, r, const r, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_EXT_RAW( name, r, r, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param)

#define BOOST_RTL_RNG_ALGO20_EXT(name) \
    BOOST_RTL_RNG_ALGO20_EXT_RAW( name, r1, const r1, r2, r2&) \
    BOOST_RTL_RNG_ALGO20_EXT_RAW( name, r1, r1, r2, r2&) \
    BOOST_RTL_RNG_ALGO20_EXT_RAW( name, r1, const r1, r2, const r2&) \
    BOOST_RTL_RNG_ALGO20_EXT_RAW( name, r1, r1, r2, const r2&)

#define BOOST_RTL_RNG_ALGO21_EXT(name, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_EXT_RAW( name, r1, const r1, r2, r2&, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_EXT_RAW( name, r1, r1, r2, r2&, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_EXT_RAW( name, r1, const r1, r2, const r2&, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_EXT_RAW( name, r1, r1, r2, const r2&, val3_type, val3_param)

#else


#define BOOST_RTL_RNG_ALGO10_EXT(name) \
    BOOST_RTL_RNG_ALGO10_EXT_RAW( name, r, r)

#define BOOST_RTL_RNG_ALGO11_EXT(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW( name, r, r, val2_type, val2_param)

#define BOOST_RTL_RNG_ALGO12_EXT(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW( name, r, r, val2_type, val2_param, val3_type, val3_param)

// especially for equal_range
#define BOOST_RTL_RNG_ALGO11_EXT2(name, val2_type, val2_param) \
    BOOST_RTL_RNG_ALGO11_EXT_RAW2( name, r, r, val2_type, val2_param)

// especially for equal_range
#define BOOST_RTL_RNG_ALGO12_EXT2(name, val2_type, val2_param, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO12_EXT_RAW2( name, r, r, val2_type, val2_param, val3_type, val3_param)

#define BOOST_RTL_RNG_ALGO13_EXT(name, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param) \
    BOOST_RTL_RNG_ALGO13_EXT_RAW( name, r, r, val2_type, val2_param, val3_type, val3_param, val4_type, val4_param)

#define BOOST_RTL_RNG_ALGO20_EXT(name) \
    BOOST_RTL_RNG_ALGO20_EXT_RAW( name, r1, r1, r2, r2&) 

#define BOOST_RTL_RNG_ALGO21_EXT(name, val3_type, val3_param) \
    BOOST_RTL_RNG_ALGO21_EXT_RAW( name, r1, r1, r2, r2&, val3_type, val3_param) 

#endif // BOOST_RTL_WORKAROUND_VC6








/* 
    Algorithms that DO NOT return a range.
*/

//template<class in_it> typename iterator_traits<in_it>::difference_type
//distance(in_it first, in_it last);
// FIXME - return difference_type
BOOST_RTL_RNG_ALGO10( distance, int)

//template <class in_it, class Distance>
//void advance(in_it& i, Distance n);
// FIXME advance should be something like
template< class c, class Distance>
void advance( crange<c> & r, Distance n) {
    typedef typename crange<c>::iterator iterator;
    iterator i = r.begin(); 
    ::std::advance(i, n);
    r.begin( i);
}


//template<class in_it, class Function>
//Function for_each(in_it first, in_it last, Function f);
BOOST_RTL_RNG_ALGO11( for_each, function, function, function)

// template<class in_it, class T> typename iterator_traits<in_it>::difference_type
// count(in_it first, in_it last, const T& value);
//     FIXME - in the future, use difference_type
BOOST_RTL_RNG_ALGO11( count, int, T, const T&)

// template<class in_it, class Predicate> typename iterator_traits<in_it>::difference_type
// count_if(in_it first, in_it last, Predicate pred);
//     FIXME - in the future, use difference_type
BOOST_RTL_RNG_ALGO11( count_if, int, T, const T&)

// 
// template<class in_it1, class in_it2> pair<in_it1, in_it2> 
// mismatch(in_it1 first1, in_it1 last1, in_it2 first2);
template< class r1, class r2>
struct mismatch_return_finder {
    typedef typename range_finder<r1>::range_type r1_range_type;
    typedef typename r1_range_type::iterator it1;
    // FIXME - if r2 is a const container, use const_iterator!!!
    typedef typename r2::iterator it2;
    typedef ::std::pair<it1, it2> return_type;
};

//FIXME - fails under VC6
//BOOST_RTL_RNG_ALGO20( mismatch, (BOOST_DEDUCED_TYPENAME mismatch_return_finder<r1,r2>::return_type))

// 
// template<class in_it1, class in_it2, class BinaryPredicate> pair<in_it1, in_it2> 
// mismatch(in_it1 first1, in_it1 last1, in_it2 first2, BinaryPredicate pred);
//FIXME - fails under VC6
//BOOST_RTL_RNG_ALGO21( mismatch, (BOOST_DEDUCED_TYPENAME mismatch_return_finder<r1,r2>::return_type), pred, pred)



// 
// template<class in_it1, class in_it2> bool 
// equal(in_it1 first1, in_it1 last1, in_it2 first2);
BOOST_RTL_RNG_ALGO20( equal, bool)

// 
// template<class in_it1, class in_it2, class BinaryPredicate> bool 
// equal(in_it1 first1, in_it1 last1, in_it2 first2, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO21( equal, bool, pred, pred)


// 
// template<class in_it, class out_it> out_it 
// copy(in_it first, in_it last, out_it result);
BOOST_RTL_RNG_ALGO11( copy, out_it, out_it, out_it)

// 
// template<class bid_it1, class bid_it2> bid_it2 
// copy_backward(bid_it1 first, bid_it1 last, bid_it2 result);
//     FIXME - here, we could actually return a range
BOOST_RTL_RNG_ALGO20( copy_backward, BOOST_DEDUCED_TYPENAME r2::iterator)

// 
// template<class fwd_it1, class fwd_it2> fwd_it2 
// swap_ranges(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2);
//     FIXME - here, we could actually return a range
BOOST_RTL_RNG_ALGO20( swap_ranges, BOOST_DEDUCED_TYPENAME r2::iterator)

// 
// template<class fwd_it1, class fwd_it2>
// void iter_swap(fwd_it1 a, fwd_it2 b);
BOOST_RTL_RNG_ALGO20_void( iter_swap)

// 
// template<class in_it, class out_it, class UnaryOperation> out_it 
// transform(in_it first, in_it last, out_it result, UnaryOperation op);
BOOST_RTL_RNG_ALGO12( transform, out_it, out_it, out_it, op, op)

// 
// template<class in_it1, class in_it2, class out_it, class BinaryOperation> out_it 
// transform(in_it1 first1, in_it1 last1, in_it2 first2, out_it result, BinaryOperation binary_op);
BOOST_RTL_RNG_ALGO22( transform, out_it, out_it, out_it, op, op)

// 
// template<class fwd_it, class T> void 
// replace(fwd_it first, fwd_it last, const T& old_value, const T& new_value);
BOOST_RTL_RNG_ALGO12_void( replace, T, const T&, T_same, const T_same&)

// 
// template<class fwd_it, class Predicate, class T> void 
// replace_if(fwd_it first, fwd_it last, Predicate pred, const T& new_value);
BOOST_RTL_RNG_ALGO12_void( replace_if, pred, pred, T, const T&)


// 
// template<class in_it, class out_it, class T> out_it 
// replace_copy(in_it first, in_it last, out_it result, const T& old_value, const T& new_value);
BOOST_RTL_RNG_ALGO13( replace_copy, out_it, out_it, out_it, T, const T&, T_same, const T_same&)

// 
// template<class Iterator, class out_it, class Predicate, class T> out_it 
// replace_copy_if(Iterator first, Iterator last, out_it result, Predicate pred, const T& new_value);
BOOST_RTL_RNG_ALGO13( replace_copy_if, out_it, out_it, out_it, pred, pred, T, const T&)

// 
// template<class fwd_it, class T> void 
// fill(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11_void( fill, T, const T&)

// 
// template<class out_it, class Size, class T> void 
// fill_n(out_it first, Size n, const T& value);
//useless - BOOST_RTL_RNG_ALGO12_void( fill_n, Size, Size, T, const T&)

// 
// template<class fwd_it, class Generator> void 
// generate(fwd_it first, fwd_it last, Generator gen);
BOOST_RTL_RNG_ALGO11_void( generate, generator, generator)


// 
// template<class out_it, class Size, class Generator> void 
// generate_n(out_it first, Size n, Generator gen);
// useless - BOOST_RTL_RNG_ALGO12_void( generate_n, Size, Size, generator, generator)

// 
// template<class in_it, class out_it> out_it 
// unique_copy(in_it first, in_it last, out_it result);
BOOST_RTL_RNG_ALGO11( unique_copy, out_it, out_it, out_it)

// 
// template<class in_it, class out_it, class BinaryPredicate> out_it 
// unique_copy(in_it first, in_it last, out_it result, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO12( unique_copy, out_it, out_it, out_it, pred, pred)

// 
// template<class bid_it> void 
// reverse(bid_it first, bid_it last);
BOOST_RTL_RNG_ALGO10_void( reverse)

// 
// template<class bid_it, class out_it> out_it 
// reverse_copy(bid_it first, bid_it last, out_it result);
BOOST_RTL_RNG_ALGO11( reverse_copy, out_it, out_it, out_it)

// 
// template<class fwd_it> void 
// rotate(fwd_it first, fwd_it middle, fwd_it last);
BOOST_RTL_RNG_ALGO11_void( rotate, fwd_it, fwd_it)

// 
// template<class fwd_it, class out_it> out_it 
// rotate_copy(fwd_it first, fwd_it middle, fwd_it last, out_it result);
BOOST_RTL_RNG_ALGO12( rotate_coy, out_it, fwd_it, fwd_it, out_it, out_it)

// 
// template<class rand_it> void 
// random_shuffle(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( random_shuffle)

// 
// template<class rand_it, class RandomNumberGenerator> void 
// random_shuffle(rand_it first, rand_it last, RandomNumberGenerator& rand);
BOOST_RTL_RNG_ALGO11_void( random_shuffle, generator, generator&)

// 
// template<class rand_it> void 
// sort(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( sort)

// 
// template<class rand_it, class Compare>
// void sort(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( sort, pred, pred)

// 
// template<class rand_it> void 
// stable_sort(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( stable_sort)

// 
// template<class rand_it, class Compare> void 
// stable_sort(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( stable_sort, pred, pred)

// 
// template<class rand_it> void 
// partial_sort(rand_it first, rand_it middle, rand_it last);
BOOST_RTL_RNG_ALGO11_void( partial_sort, rand_it, rand_it)

// 
// template<class rand_it, class Compare> void 
// partial_sort(rand_it first, rand_it middle, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO12_void( partial_sort, rand_it, rand_it, pred, pred)

// 
// template<class rand_it> void 
// nth_element(rand_it first, rand_it nth, rand_it last);
BOOST_RTL_RNG_ALGO11_void( nth_element, rand_it, rand_it)

// 
// template<class rand_it, class Compare> void 
// nth_element(rand_it first, rand_it nth, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO12_void( nth_element, rand_it, rand_it, pred, pred)

// 
// template<class fwd_it, class T> bool 
// binary_search(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11( binary_search, bool, T, const T&)

// 
// template<class fwd_it, class T, class Compare> bool 
// binary_search(fwd_it first, fwd_it last, const T& value, Compare comp);
BOOST_RTL_RNG_ALGO12( binary_search, bool, T, const T&, pred, pred)

// 
// template<class in_it1, class in_it2, class out_it> out_it 
// merge(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result);
BOOST_RTL_RNG_ALGO21( merge, out_it, out_it, out_it)

// 
// template<class in_it1, class in_it2, class out_it, class Compare> out_it 
// merge(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result, Compare comp);
BOOST_RTL_RNG_ALGO22( merge, out_it, out_it, out_it, pred, pred)

// 
// template<class bid_it> void 
// inplace_merge(bid_it first, bid_it middle, bid_it last);
BOOST_RTL_RNG_ALGO11_void( inplace_merge, bid_it, bid_it)

// 
// template<class bid_it, class Compare> void 
// inplace_merge(bid_it first, bid_it middle, bid_it last, Compare comp);
BOOST_RTL_RNG_ALGO12_void( inplace_merge, bid_it, bid_it, pred, pred)


// 
// template<class in_it1, class in_it2> bool 
// includes(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2);
BOOST_RTL_RNG_ALGO20( includes, bool)

// 
// template<class in_it1, class in_it2, class Compare> bool 
// includes(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, Compare comp);
BOOST_RTL_RNG_ALGO21( includes, bool, comp, comp)

// 
// template<class in_it1, class in_it2, class out_it> out_it 
// set_union(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result);
BOOST_RTL_RNG_ALGO21( set_union, out_it, out_it, out_it)

// 
// template<class in_it1, class in_it2, class out_it, class Compare>
// out_it set_union(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result, Compare comp);
BOOST_RTL_RNG_ALGO22( set_union, out_it, out_it, out_it, pred, pred)

// 
// template<class in_it1, class in_it2, class out_it> out_it 
// set_intersection(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result);
BOOST_RTL_RNG_ALGO21( set_intersection, out_it, out_it, out_it)

// 
// template<class in_it1, class in_it2, class out_it, class Compare> out_it 
// set_intersection(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result, Compare comp);
BOOST_RTL_RNG_ALGO22( set_intersection, out_it, out_it, out_it, pred, pred)

// 
// template<class in_it1, class in_it2, class out_it> out_it 
// set_difference(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result);
BOOST_RTL_RNG_ALGO21( set_difference, out_it, out_it, out_it)

// 
// template<class in_it1, class in_it2, class out_it, class Compare> out_it 
// set_difference(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result, Compare comp);
BOOST_RTL_RNG_ALGO22( set_difference, out_it, out_it, out_it, pred, pred)

// 
// template<class in_it1, class in_it2, class out_it> out_it
// set_symmetric_difference(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result);
BOOST_RTL_RNG_ALGO21( set_symmetric_difference, out_it, out_it, out_it)

// 
// template<class in_it1, class in_it2, class out_it, class Compare> out_it
// set_symmetric_difference(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, out_it result, Compare comp);
BOOST_RTL_RNG_ALGO22( set_symmetric_difference, out_it, out_it, out_it, pred, pred)

// 
// 
// template<class rand_it> void 
// push_heap(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( push_heap)

// 
// template<class rand_it, class Compare> void 
// push_heap(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( push_heap, pred, pred)

// 
// template<class rand_it> void 
// pop_heap(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( pop_heap)

// 
// template<class rand_it, class Compare> void 
// pop_heap(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( pop_heap, pred, pred)

// 
// template<class rand_it> void 
// make_heap(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( make_heap)

// 
// template<class rand_it, class Compare> void 
// make_heap(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( make_heap, pred, pred)

// 
// template<class rand_it> void 
// sort_heap(rand_it first, rand_it last);
BOOST_RTL_RNG_ALGO10_void( sort_heap)

// 
// template<class rand_it, class Compare> void 
// sort_heap(rand_it first, rand_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_void( sort_heap, pred, pred)

// 
// template<class in_it1, class in_it2> bool 
// lexicographical_compare(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2);
BOOST_RTL_RNG_ALGO20( lexicographical_compare, bool)

// 
// template<class in_it1, class in_it2, class Compare> bool 
// lexicographical_compare(in_it1 first1, in_it1 last1, in_it2 first2, in_it2 last2, Compare comp);
BOOST_RTL_RNG_ALGO21( lexicographical_compare, bool, pred, pred)

// 
// template<class bid_it> bool 
// next_permutation(bid_it first, bid_it last);
BOOST_RTL_RNG_ALGO10( next_permutation, bool)

// 
// template<class bid_it, class Compare> bool 
// next_permutation(bid_it first, bid_it last, Compare comp);
BOOST_RTL_RNG_ALGO11( next_permutation, bool, pred, pred)

// 
// template<class bid_it> bool 
// prev_permutation(bid_it first, bid_it last);
BOOST_RTL_RNG_ALGO10( prev_permutation, bool)

// 
// template<class bid_it, class Compare> bool 
// prev_permutation(bid_it first, bid_it last, Compare comp);
BOOST_RTL_RNG_ALGO11( prev_permutation, bool, pred, pred)

// template<class in_it, class out_it, class T> out_it 
// remove_copy(in_it first, in_it last, out_it result, const T& value);
BOOST_RTL_RNG_ALGO12( remove_copy, out_it, out_it, out_it, T, const T&)

// 
// template<class in_it, class out_it, class Predicate> out_it 
// remove_copy_if(in_it first, in_it last, out_it result, Predicate pred);
BOOST_RTL_RNG_ALGO12( remove_copy_if, out_it, out_it, out_it, pred, pred)





// 
// 
// 
// 
// * These functions returns a range

/*
    Right now, we always return a range, from the element "found",
    up to the end of the original range.

    This might not necessary be what you want.

    Therefore, FIXME (to do).

    Allow for algorithm of this type:
    find_if<what_to_return>(range);

    What to return is any of:
    1. return the iterator itself ("r_it")
    2. (default - as is now) return the range from the 
       found iterator up to the end ("r_to_end")
    3. return the range starting from begin, up to the found iterator ("r_from_beg")

    Note: see end of file.
*/
// template<class in_it, class T> in_it 
// find(in_it first, in_it last, const T& value);
BOOST_RTL_RNG_ALGO11_EXT( find, T, const T&)

// 
// template<class in_it, class Predicate> in_it 
// find_if(in_it first, in_it last, Predicate pred);
BOOST_RTL_RNG_ALGO11_EXT( find_if, pred, pred)

// 
// template<class fwd_it1, class fwd_it2> fwd_it1 
// find_end(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2);
BOOST_RTL_RNG_ALGO20_EXT( find_end)

// 
// template<class fwd_it1, class fwd_it2, class BinaryPredicate> fwd_it1
// find_end(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO21_EXT( find_end, pred, pred)

// 
// template<class fwd_it1, class fwd_it2> fwd_it1
// find_first_of(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2);
BOOST_RTL_RNG_ALGO20_EXT( find_first_of)

// 
// template<class fwd_it1, class fwd_it2, class BinaryPredicate> fwd_it1
// find_first_of(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO21_EXT( find_first_of, pred, pred)

// 
// template<class fwd_it> fwd_it 
// adjacent_find(fwd_it first, fwd_it last);
BOOST_RTL_RNG_ALGO10_EXT( adjacent_find)

// template<class fwd_it, class BinaryPredicate> fwd_it 
// adjacent_find(fwd_it first, fwd_it last, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO11_EXT( adjacent_find, pred, pred)

// 
// 
// template<class fwd_it1, class fwd_it2> fwd_it1 
// search(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2);
BOOST_RTL_RNG_ALGO20_EXT( search)

// 
// template<class fwd_it1, class fwd_it2, class BinaryPredicate> fwd_it1 
// search(fwd_it1 first1, fwd_it1 last1, fwd_it2 first2, fwd_it2 last2, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO21_EXT( search, pred, pred)

// 
// template<class fwd_it, class Size, class T> fwd_it 
// search_n(fwd_it first, fwd_it last, Size count, const T& value);
BOOST_RTL_RNG_ALGO12_EXT( search_n, Size, Size, T, const T&)

// 
// template<class fwd_it, class Size, class T, class BinaryPredicate> fwd_it1 
// search_n(fwd_it first, fwd_it last, Size count, const T& value, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO13_EXT( search_n, Size, Size, T, const T&, pred, pred)

// 
// ***** return the range of elements to be removed;
//        use them in correlation with r_erase.
//

// template<class fwd_it, class T> fwd_it 
// remove(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11_EXT( remove, T, const T&)

// 
// template<class fwd_it, class Predicate> fwd_it 
// remove_if(fwd_it first, fwd_it last, Predicate pred);
BOOST_RTL_RNG_ALGO11_EXT( remove_if, pred, pred)

// 
// template<class fwd_it> fwd_it 
// unique(fwd_it first, fwd_it last);
BOOST_RTL_RNG_ALGO10_EXT( unique)

// 
// template<class fwd_it, class BinaryPredicate> fwd_it 
// unique(fwd_it first, fwd_it last, BinaryPredicate pred);
BOOST_RTL_RNG_ALGO11_EXT( unique, pred, pred)

// 
// 
// // return the range that satisfies the predicate
// template<class bid_it, class Predicate> bid_it 
// partition(bid_it first, bid_it last, Predicate pred);
BOOST_RTL_RNG_ALGO11_EXT( partition, pred, pred)

// 
// template<class bid_it, class Predicate> bid_it 
// stable_partition(bid_it first, bid_it last, Predicate pred);
BOOST_RTL_RNG_ALGO11_EXT( stable_partition, pred, pred)

// 
// template<class in_it, class rand_it> rand_it
// partial_sort_copy(in_it first, in_it last, rand_it result_first, rand_it result_last);
// FIXME - watch out what we return!!!

// 
// template<class in_it, class rand_it, class Compare> rand_it
// partial_sort_copy(in_it first, in_it last, rand_it result_first, rand_it result_last, Compare comp);
// FIXME - watch out what we return!!!


// 
// template<class fwd_it, class T> fwd_it 
// lower_bound(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11_EXT( lower_bound, T, const T&)

// 
// template<class fwd_it, class T, class Compare> fwd_it 
// lower_bound(fwd_it first, fwd_it last, const T& value, Compare comp);
BOOST_RTL_RNG_ALGO12_EXT( lower_bound, T, const T&, pred, pred)

// 
// template<class fwd_it, class T> fwd_it 
// upper_bound(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11_EXT( upper_bound, T, const T&)

// 
// template<class fwd_it, class T, class Compare> fwd_it 
// upper_bound(fwd_it first, fwd_it last, const T& value, Compare comp);
BOOST_RTL_RNG_ALGO12_EXT( upper_bound, T, const T&, pred, pred)

// 
// template<class fwd_it, class T> pair<fwd_it, fwd_it>
// equal_range(fwd_it first, fwd_it last, const T& value);
BOOST_RTL_RNG_ALGO11_EXT2( equal_range, T, const T&)


// 
// template<class fwd_it, class T, class Compare> pair<fwd_it, fwd_it>
// equal_range(fwd_it first, fwd_it last, const T& value, Compare comp);
BOOST_RTL_RNG_ALGO12_EXT2( equal_range, T, const T&, pred, pred)


// 
// template<class fwd_it> fwd_it 
// min_element(fwd_it first, fwd_it last);
BOOST_RTL_RNG_ALGO10_EXT( min_element)

// 
// template<class fwd_it, class Compare> fwd_it 
// min_element(fwd_it first, fwd_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_EXT( min_element, pred, pred)

// 
// template<class fwd_it> fwd_it 
// max_element(fwd_it first, fwd_it last);
BOOST_RTL_RNG_ALGO10_EXT( max_element)

// 
// template<class fwd_it, class Compare> fwd_it 
// max_element(fwd_it first, fwd_it last, Compare comp);
BOOST_RTL_RNG_ALGO11_EXT( max_element, pred, pred)






} // namespace rng
}}

#endif
/*
Compiles under VC6 !!!

#include <iostream>
typedef enum range_return_type {
    it,
    from_beg,
    to_end
};


template< class T>
void test( const T& val){
    std::cout << "general" << std::endl;
}

template< range_return_type r, class T>
void test( const T& val, range_return_type ignore_vc_bug = r){
    std::cout << "specific" << std::endl;
}

struct try_it {};
int main(int argc, char* argv[])
{
    test( 1);
    test( try_it() );
    test<it>( 1 );
    test<from_beg>( 2 );
    test<it>( try_it() );
    test<to_end>( 2 );
}
*/
